"""
Converts the 5' and 3' coordinates of reads into coordinates on meta-features.
In essence, each feature is divided into bins; the 5' and 3' coordinates of
each read contained in each feature are replaced by the bin index in which
they are included. For example for a feature of length 100 containing a read
with coordinates [5', 3']: [15, 95] the corresponding meta-coordinates for
10 bins are [1, 9]. THe script works only for +ve strands and BED coordinates
grater than certain length of nucleotide set in 'min-len' parameter.
"""

import sys
import pysam
import argparse

def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-i", "--ifile",
                    help="Input SAM/BAM file")
    parser.add_argument("-b", "--bed",
                    help="Bed file with gene feature coordinates")
    parser.add_argument("-f", "--sam", action='store_true',
                    help="Use this option if input file SAM")
    parser.add_argument("-c", "--bins", default=20, type=int,
                    help="Number of bins, per feature, default: %(default)s" )
    parser.add_argument("-l", "--min-len", default=200, type=int,
                    help="Minimum transcript length processed, default:%(default)s")
    parser.add_argument("-d", "--col-qname-name", default="qname",
                    help="Header for column query name, default: qname")
    parser.add_argument("-e", "--col-feat-name", default="feat",
                    help="Header of column wih features, default: feat")
    parser.add_argument("-g", "--col-bin5p-name", default="bin5p",
                    help="Header for 5 prime bins of read, default: bin5p")
    parser.add_argument("-j", "--col-bin3p-name", default="bin3p",
                    help="Header for 3 prime bins of read, default: bin3p")
    parser.add_argument("-s", "--col_delimiter", default="\t",
                    help="Delimiter of output file columns, default: tab(\t)")
    args = parser.parse_args()

    filemode = "rb"
    if args.sam:
        filemode = "r"

    region_coords = {}
    with open(args.bed) as bedfile:
        for line in bedfile:
            line = line.strip().split("\t")
            if line[5] == "-":
                print("Error: Code does not support calculations for -ve strand")
                sys.exit(1)
            refid = line[0]
            ref_start = int(line[1])
            ref_end = int(line[2])
            if (ref_end - ref_start) < args.min_len:
                continue
            if refid in region_coords:
                print("Error: Dulplicate transcripts found ")
                sys.exit(1)
            region_coords[refid] = [ref_start, ref_end]

    delim = args.col_delimiter
    print(delim.join([args.col_qname_name, args.col_feat_name, args.col_bin5p_name, args.col_bin3p_name]))

    infile = pysam.AlignmentFile(args.ifile, filemode)
    for read in infile:
        if read.is_unmapped or read.is_reverse: #check for unmapped and -ve strand
            continue
        qname = read.query_name
        refname = read.reference_name
        qstart = read.reference_start
        qstop = read.reference_end - 1

        if refname in region_coords:
            bed_start = region_coords[refname][0]
            bed_end = region_coords[refname][1]

            bin_len = (bed_end - bed_start)/args.bins
            bin5p = int((qstart - bed_start)/bin_len)
            bin3p = int((qstop - bed_start)/bin_len)

            if bin5p < 0 or bin5p >= args.bins:
                bin5p = "NA"

            if bin3p < 0 or bin3p >= args.bins:
                bin3p = "NA"
            print(delim.join([qname, refname, str(bin5p), str(bin3p)]))



"""
Plots the distribution of reads in a BAM file around a given set of features
in a BED file. The user can specify which end (5' or 3') of the reads and the
features will be used as reference for the comparison. For example: We assume
that the user selects the 5' end of the reads and the 5' end of the features
as reference. Then a read that maps at position 10 of chr1 will be at a
relative position of -5 nt compared to a feature aligning at position 15 of
chr1. The same concept is applied for all reads against all features and a
distribution of relative positions is constructed.
"""

import pysam
import argparse
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

def find_pos(start, end, strand, pos):
    if pos != '5p' and pos != '3p':
        raise ValueError('Incorrectly specified position')
    if strand != '+' and strand != '-':
        raise ValueError('Incorrectly specified strand')
    final_pos = start
    if strand == '-' and pos == '5p' or strand == '+' and pos == '3p':
        final_pos = end
    return final_pos

def relative_pos(bed, read, fpos, rpos):
    """
    rel_pos calculates the relative position between the two reference
    positions of a bed and a reads entry.
    """
    # NOTE: the relative position is supposed to indicate if one is upstream
    # or downstream of the other. Pay attention to the strand.
    fstart, fend, fstrand = bed[0], bed[1], bed[2]
    feat_pos = find_pos(fstart, fend, fstrand, fpos)
    read_strand = '+' if read.is_forward else '-'
    read_pos = find_pos(read.reference_start, read.reference_end - 1, read_strand, rpos)
    rel_pos = read_pos - feat_pos
    if fstrand == '-':
        rel_pos *= -1
    return rel_pos

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-m","--bam",
                    help = "BAM file with reads. Must be indexed.")
    parser.add_argument("-b","--bed",
                    help = "BED file with features.")
    parser.add_argument("-u","--up", type = int, default = 100,
                    help = "Number of nts to plot upstream of pos. (default: %(default)s)")
    parser.add_argument("-d","--down", type = int, default = 100,
                    help = "Number of nts to plot downstream of pos. (default: %(default)s)")
    parser.add_argument("-f","--fpos", default='5p',
                    help = "Reference point for features; one of 5p or 3p (default: %(default)s)")
    parser.add_argument("-r","--rpos", default='5p',
                    help = "Reference point for reads; one of 5p or 3p (default: %(default)s)")
    parser.add_argument("-o", "--pdf",
                    help = "Output pdf file with plot")
    args = parser.parse_args()


    # Initialize a histogram
    hist = {i : 0 for i in range(-1 * args.up, args.down + 1)}

    # Open the bam file
    bamfile = pysam.AlignmentFile(args.bam, "rb")


    # Open and loop on the BED file line by line. For each line query the BAM file
    # to get overlapping reads. Calculate relative positions and add in the
    # histogram. NOTE: BED files are 0 based files.
    feature_count = 0
    total_counts = 0
    allowed_counts = 0
    with open(args.bed) as bed:
        for line in bed:
            feature_count += 1
            cols = line.strip().split('\t')
            chr = cols[0]
            start = int(cols[1])
            end = int(cols[2]) - 1 # BED files are zero based i.e. [coord,coord)
            strand = str(cols[5])  # Can be + or -
            feat = [start, end, strand]
            # Fetch overlapping BAM entries
            reads = bamfile.fetch(chr, start, end)
            #reads_counts.append(bamfile.count(chr, start, end))
            # Calculate relative positions and append to histogram
            for read in reads:
                total_counts += 1
                rel_pos = relative_pos(feat, read, args.fpos, args.rpos)
                if rel_pos > -1 * args.up and rel_pos < args.down:
                    allowed_counts += 1
                    hist[rel_pos] += 1

    positions = [int(x) for x in hist.keys()]
    positions.sort()
    values = [hist[i] for i in positions]
    # print('Total features: ' + str(feature_count))
    # print('Total reads overlap: ' + str(sum(total_counts)))
    # print('Total reads allowed: ' + str(allowed_counts))

    with PdfPages(args.pdf) as pages:
        fig, ax = plt.subplots(layout='constrained')
        plt.bar(positions, values)
        plt.xlabel('Relative position to features ' + args.fpos + ' end')
        plt.ylabel('Read count')
        pages.savefig(fig)
        plt.close()


if __name__ == "__main__":
    main()
